"""
Code for rotations of points
"""

import numpy as np
import SimpleITK as sitk
from scipy.spatial.transform import Rotation

from . import utils as ut


def define_euler_rotation(rx, ry, rz, degrees=True, order="xyz"):
    """
    Wrapper of scipy.spatial.transform.Rotation.from_euler

    Parameters
    ----------
    rx : Float
        Angle to rotate about X
    ry : Float
        Angle to rotate about Y
    rz : Float
        Angle to rotate about Z
    degrees : Bool, optional
        Are the rotations in degrees?. The default is True.
    order: string, optional
        Order of axes to transform as string. Default is 'xyz',
        meaning transform will happen x-->y-->z

    Returns
    -------
    Scipy 3d rotation
        scipy 3.

    """
    return Rotation.from_euler(order, [rx, ry, rz], degrees=True)


def rotate_about_and_translate(points, rotation, pivot, translation):
    """
    Rotates points about a pivot point,
    then apply translation (add the translation values)


    Parameters
    ----------
    points : (Nx3) numpy array
        Points to rotate. Each point gets its own row.
    rototation : Scipy `Rotation` object
        use `define_euler_rotation` or
        `scipy.spatial.transform.Rotation` constructor to create
    pivot : (1x3) numpy array
        Point to rotate around
    translation: (1x3) numpy array
        Additional translation to apply to points


    Returns
    -------
    (Nx3) numpy array
        Rotated points

    """
    return rotate_about(points, rotation, pivot) + translation


def rotate_about(points, rotation, pivot):
    """
    Rotates points about a pivot point

    Parameters
    ----------
    points : (Nx3) numpy array
        Points to rotate. Each point gets its own row.
    rototation : Scipy `Rotation` object
        use `define_euler_rotation` or
        `scipy.spatial.transform.Rotation` constructor to create
    pivot : (1x3) numpy array
        Point to rotate around

    Returns
    -------
    (Nx3) numpy array
        Rotated points

    """
    return rotation.apply(points - pivot) + pivot


def rotation_matrix_to_sitk(
    rotation, center=np.array((0, 0, 0)), translation=np.array((0, 0, 0))
):
    """Convert numpy array rotation matrix to sitk affine

    Parameters
    ----------
    rotation : np.ndarray (3 x 3)
        matrix representing rotation matrix in three dimensions
    center : np.ndarray (3)
        vector representing center of rotation, default is origin
    translation : np.ndarray (3)
        vector representing translation of transform (after rotation), default
        is zero

    Returns
    -------
    SITK transform
        with parameters matching the input object

    """
    S = sitk.AffineTransform(3)
    S.SetMatrix(tuple(rotation.flatten()))
    S.SetTranslation(translation.tolist())
    S.SetCenter(center.tolist())
    return S


def scipy_rotation_to_sitk(
    rotation, center=np.array((0, 0, 0)), translation=np.array((0, 0, 0))
):
    """
    Convert Scipy 'Rotation' object to equivalent sitk

    Parameters
    ----------
    rotation : Scipy `Rotation` object
        use `define_euler_rotation` or
        `scipy.spatial.transform.Rotation` constructor to create

    Returns
    -------
    SITK transform
        with parameters matching the input object

    """
    S = rotation_matrix_to_sitk(rotation.as_matrix(), center, translation)
    return S


def rotation_matrix_from_vectors(a, b):
    """Find rotation matrix to align a with b


    Parameters
    ----------
    a : np.ndarray (N)
        vector to be aligned with b
    b : np.ndarray (N)
        vector

    Returns
    -------
    rmat : np.ndarray (NxN)
        Rotation matrix such that `rmat @ a` is parallel to `b`
    """
    # Follows Rodrigues` rotation formula
    # https://math.stackexchange.com/a/476311

    nd = a.shape[0]
    if nd != b.shape[0]:
        raise ValueError("a must be same size as b")
    na = ut.norm_vec(a)
    nb = ut.norm_vec(b)
    c = np.dot(na, nb)
    if c == -1:
        return -np.eye(nd)
    v = np.cross(na, nb)
    ax = ut.skew_symmetric_cross_product_matrix(v)
    rotmat = np.eye(nd) + ax + ax @ ax * (1 / (1 + c))
    return rotmat



"""
Code to handle sitk volume loading and rotating

SimpleITK example code is under Apache License, see:
https://github.com/SimpleITK/TUTORIAL/blob/main/LICENSE

"""

import itertools as itr

import numpy as np
import SimpleITK as sitk

from . import utils as ut


def resample(
    image,
    transform=None,
    output_spacing=None,
    output_origin=None,
    output_size=None,
    interpolator=sitk.sitkLinear,
):
    """
    Wrapper to generically handle SimpleITK resampling on different image
    matrices. Includes optional application of a transform.
    Only 3d is currently implemented.

    Code is modified from the 2d example in.
    https://simpleitk.org/SPIE2018_COURSE/images_and_resampling.pdf
    and
    https://github.com/SimpleITK/TUTORIAL/blob/main/...
        02_images_and_resampling.ipynb

    Parameters
    ----------
    image : SimpleITK image
        image to transform.
    transform : SimpleITK Affine Transform, optional
        If no transform is passed, use a identity transform matrix
    output_spacing : (Nx1) array, optional
        If not passed, copies from image
    output_origin : (Nx1) array, optional
        If not passed, copies from image
    output_size : (Nx1) array, optional
        If not passed, computes automatically to fully encompass
        transformed image.
    interpolator: SimpleITK Interpolator, optional
        If not passed, defaults to sitk.sitkLinear
        See sitk documentation for options.

    Returns
    -------
    resampled_image : SimpleITK image
        resampled image with transform applied.

    """
    if len(image.GetSize()) == 3:
        return resample3D(
            image,
            transform=transform,
            output_spacing=output_spacing,
            output_origin=output_origin,
            output_size=output_size,
            interpolator=interpolator,
        )
    else:
        raise NotImplementedError(
            "Resample currently only supports 3D transformations"
        )


def resample3D(
    image,
    transform=None,
    output_spacing=None,
    output_origin=None,
    output_size=None,
    interpolator=sitk.sitkLinear,
):
    """
    Resample a 3D sitk image, with the option to apply a transform

    Parameters
    ----------
    image : SimpleITK image
        image to transform.
    transform : SimpleITK Affine Transform, optional
        If no transform is passed, use a identity transform matrix
    output_spacing : (3x1) array, optional
        If not passed, copies from image
    output_origin : (3x1) array, optional
        If not passed, copies from image
    output_size : (3x1) array, optional
        If not passed, computes automatically to fully encompass
        transformed image.

    Returns
    -------
    resampled_image : SimpleITK image
        resampled image with transform applied.

    """
    if transform is None:
        transform = sitk.AffineTransform(3)

    inv_transform = transform.GetInverse()
    extrema_transformed = list(
        map(
            lambda x: inv_transform.TransformPoint(  # Apply inverse transform
                image.TransformIndexToPhysicalPoint(x)  # To the physical point
            ),
            itr.product(
                *map(lambda x: (0, x), image.GetSize())
            ),  # for all pairs of extreme indices
        )
    )

    extrema_arr = np.vstack(extrema_transformed)
    minmax = np.vstack(
        list(map(lambda x: x(extrema_arr, axis=0), [np.min, np.max]))
    )

    #
    if output_spacing is None:
        spacing = np.empty(3)
        spacing.fill(np.median(np.array(image.GetSpacing())))
        output_spacing = tuple(spacing)

    if output_origin is None:
        output_origin = minmax[0, :].tolist()

    # Compute grid size based on the physical size and spacing.
    if output_size is None:
        output_size = (
            np.round(np.diff(minmax, axis=0).squeeze() / spacing)
            .astype(int)
            .tolist()
        )

    resampled_image = sitk.Resample(
        image,
        output_size,
        transform,
        interpolator,
        output_origin,
        output_spacing,
        tuple(np.eye(3).flatten()),
    )
    return resampled_image


def transform_sitk_indices_to_physical_points(simage, index_arr):
    """Transforms indices indices of simage to physical points

    For a SimpleITK image `simage` and a list of indices `index_arr`, transform
    each index to a physical point.

    Parameters
    ----------
    simage : M-d SimpleITK image
    index_arr : numpy.ndarray (NxM)
        matrix of indices of `simage`, where each row is an index

    Returns
    -------
    position_arr: numpy.ndarray (NxM)
        matrix of physical points for each index in `index_arr`
    """
    position_arr = np.zeros_like(index_arr, dtype="float32")
    npt = index_arr.shape[0]
    for ptno in range(npt):
        ndx = tuple(map(lambda x: x.item(), index_arr[ptno, :]))
        position_arr[ptno, :] = simage.TransformContinuousIndexToPhysicalPoint(
            ndx
        )
    return position_arr


def find_points_equal_to(simage, val):
    """Find physical points in simage equal to val
    Results are in LPS
    """
    arr = sitk.GetArrayViewFromImage(simage)
    ndxs = ut.find_indices_equal_to(arr, val)
    ndxs_xyz = ndxs[
        :, [2, 1, 0]
    ]  # convert between numpy and simpleITK indexing
    return transform_sitk_indices_to_physical_points(simage, ndxs_xyz)



"""
Code to find the rotation matrix to align a headframe to a set of holes.
"""

import itertools as itr

import numpy as np
import SimpleITK as sitk
from scipy.spatial.transform import Rotation

from . import rotations as rot
from . import sitk_volume as sv
from . import utils as ut

lps_axes = dict(
    ap=np.array([0, 1, 0]), dv=np.array([0, 0, 1]), ml=np.array([1, 0, 0])
)
def_orient_names = ("vertical", "horizontal")
def_ap_names = ("anterior", "posterior")
def_orient_comparison_axes = dict(
    horizontal=lps_axes["dv"], vertical=lps_axes["ap"]
)
def_orient_axes_dict = {
    orient: lps_axes[direction]
    for orient, direction in dict(horizontal="ap", vertical="dv").items()
}
def_design_centers = dict(
    horizontal=dict(
        anterior=np.array([-6.34, np.nan, 2.5]),
        posterior=np.array([-5.04, np.nan, 1]),
    ),
    vertical=dict(
        anterior=np.array([-5.09, 3.209, np.nan]),
        posterior=np.array([-6.84, 9.909, np.nan]),
    ),
)
def_orient_indices = dict(horizontal=[0, 2], vertical=[0, 1])
def_hole_order = dict(
    horizontal=["anterior", "posterior"],
    vertical=["posterior", "anterior"],
)


def get_segmentation_pca(seg_img, seg_vals):
    """Finds first pca axis of segmentation for segments in set seg_vals

    For each value in seg_vals, this will find the indices of seg_arr equal to
    that value and de-mean it. The first PC will then be found of the
    concatenated centered groups of indices.

    Parameters
    ---------
    seg_arr : SimpleITK.Image
        Array with annotation values in each index
    seg_vals : iterable
        Set of values that each element of seg_arr will be compared to.

    Returns
    -------
    pc_axis : first pc axis of the indices separately centered for each
    """
    # Centers each segmentation value separately
    centered = []
    for seg_val in seg_vals:
        p = sv.find_points_equal_to(seg_img, seg_val)
        m = np.mean(p, axis=0)
        centered.append(p - m)
    gp = np.concatenate(centered, axis=0)
    return ut.get_first_pca_axis(gp)


def slices_centers_of_mass(
    img, seg_img, axis_dim, seg_val, slice_seg_thresh=1
):
    """Finds the center of mass of image slices along array dimension

    Iterates through `img` along dimension `axis_dim`, finds how many elements
    of `seg_img` are equal to `seg_val`, and if that number is greater than
    or equal to `slice_seg_thresh` calculates the center of mass of the masked
    `img` on that slice. Centers of mass are based on the physical points
    corresponding to each index in `seg_img` found with
    `transform_stik_indices_to_physical_points`.

    Parameters
    ----------
    img : SimpleITK.Image
        Grayscale image used to calculate center of mass
    seg_img : SimpleITK.Image
        segmentation image used to select elements of `img`. The spatial
        information of `seg_img` will be used to determine where each element
        is in space.
    axis_dim : integer
        Axis along which `img` is sliced, in SimpleITK axis order
    seg_val : number
        elements of `img` will only be included in the center of mass if the
        corresponding element of `seg_img` is equal to `seg_val`
    slice_seg_thresh : integer
        center of mass along slices of `img` will only be calculated if at
        least `slice_seg_thresh` or more elements of `seg_img` are equal to
        `seg_val`. Default = 1.

    Returns
    -------
    com : np.ndarray (N x 3)
        center of mass for each slice of `img` meeting the criteria described
        above.
    """
    seg_arr = sitk.GetArrayViewFromImage(seg_img)
    arr = sitk.GetArrayViewFromImage(img)
    ndxs = ut.find_indices_equal_to(seg_arr, seg_val)
    ndxs_sitk = ndxs[:, ::-1]
    slice_ndxs = np.unique(ndxs_sitk[:, axis_dim])
    nmask_in_slice = np.array(
        [np.count_nonzero(ndxs_sitk[:, axis_dim] == x) for x in slice_ndxs]
    )
    sel_slice_ndxs = slice_ndxs[nmask_in_slice >= slice_seg_thresh]
    ndx_points = sv.transform_sitk_indices_to_physical_points(
        seg_img, ndxs_sitk
    )
    com = np.zeros((sel_slice_ndxs.size, 3))
    for i, slice_ndx in enumerate(sel_slice_ndxs):
        mask = ndxs_sitk[:, axis_dim] == slice_ndx
        np_ndx = tuple(ndxs[mask, :].T)
        sel_v = arr[np_ndx]
        com[i, :] = np.sum(
            sel_v[:, np.newaxis] * ndx_points[mask, :], axis=0
        ) / np.sum(sel_v)
    return com


def find_hole(img, seg_img, seg_val, sel_ndxs):
    """Find the center of a hole based on its segmentation value

    sel_ndxs is in sitk axis order!
    Returns sitk axis order
    """
    if seg_img.GetSize() != img.GetSize():
        raise ValueError("Image and segmentation must have the same shape")
    seg_arr = sitk.GetArrayViewFromImage(seg_img)
    arr = sitk.GetArrayViewFromImage(img)
    ndxs = ut.find_indices_equal_to(seg_arr, seg_val)
    if np.size(ndxs) == 0:
        return None
    ndx_points = sv.transform_sitk_indices_to_physical_points(
        seg_img, ndxs[:, [2, 1, 0]]  # convert to sitk axis order
    )
    np_ndx = tuple(ndxs.T)
    sel_v = arr[np_ndx]
    sum_sel = np.sum(sel_v)
    if sum_sel > 0:
        found_center = np.nan * np.ones(3)
        found_center[sel_ndxs] = (
            np.sum(sel_v[:, np.newaxis] * ndx_points, axis=0) / sum_sel
        )[sel_ndxs]
    else:
        return None
    return found_center


def find_holes_by_orientation(
    img,
    seg_img,
    seg_vals_dict,
    orient_indices=def_orient_indices,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
):
    """
    Find holes in an image by different orientations and anterior-posterior
    names.

    Parameters
    ----------
    img : ndarray
        The original image in which holes need to be found.
    seg_img : ndarray
        The segmented image that identifies different regions.
    seg_vals_dict : dict
        A dictionary where keys are orientation names and values are
        dictionaries.  These inner dictionaries map anterior-posterior names to
        segmentation values.
    orient_indices : dict, optional
        A dictionary where keys are orientation names and values are the
        indices used to find the holes in the image.
    orient_names : list of str, optional
        A list of orientation names.
    ap_names : list of str, optional
        A list of anterior-posterior names.

    Returns
    -------
    found_centers : dict
        A dictionary where keys are orientation names and values are
        dictionaries.  These inner dictionaries map anterior-posterior names to
        the centers of found holes.  If a hole is not found for a given
        orientation and anterior-posterior name, that entry will be missing.

    Notes
    -----
    This function iterates over the given orientations and anterior-posterior
    names to find the holes in the image using the `find_hole` function. The
    centers of the found holes are returned in a nested dictionary structure.
    """
    found_centers = {orient: dict() for orient in orient_names}
    for orient in orient_names:
        for ap in ap_names:
            if ap in seg_vals_dict[orient]:
                maybe_hole = find_hole(
                    img,
                    seg_img,
                    seg_vals_dict[orient][ap],
                    orient_indices[orient],
                )
                if maybe_hole is not None:
                    found_centers[orient][ap] = maybe_hole
    return found_centers


def find_hole_angles(
    centers_dict,
    hole_order=def_hole_order,
    orient_comparison_axis=def_orient_comparison_axes,
    orient_axis_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
):
    """
    Calculate angles between holes for each orientation.

    Parameters
    ----------
    centers_dict : dict
        Dictionary of hole centers for each orientation and anterior-posterior
        name.
    hole_order : dict, optional
        Dictionary defining the order of holes for each orientation, by default
        `def_hole_order`.
    orient_comparison_axis : dict, optional
        Dictionary of comparison axes for each orientation, by default
        `def_orient_comparison_axes`.
    orient_axis_dict : dict, optional
        Dictionary of axes for each orientation, by default
        `def_orient_axes_dict`.
    orient_names : list of str, optional
        List of orientation names, by default `def_orient_names`.

    Returns
    -------
    centers_ang : dict
        Dictionary of calculated angles for each orientation.
    """
    centers_ang = dict()
    for orient in orient_names:
        centers_diff = (
            centers_dict[orient][hole_order[orient][0]]
            - centers_dict[orient][hole_order[orient][1]]
        )
        cd_nnan = centers_diff.copy()
        cd_nnan[np.isnan(cd_nnan)] = 0
        centers_ang[orient] = ut.signed_angle_rh(
            orient_comparison_axis[orient],
            cd_nnan,
            orient_axis_dict[orient],
        )
    return centers_ang


def estimate_hole_axis_from_segmentation(seg_img, seg_vals, reference_axis):
    """
    Estimate the axis of a hole from segmentation values.

    Parameters
    ----------
    seg_img : SimpleITK.Image
        The segmentation image.
    seg_vals : list
        List of segmentation values used to identify the hole.
    reference_axis : ndarray
        Reference axis vector, direction of returned axis will be flipped if
        the dot product between reference_axis and the found axis is negative

    Returns
    -------
    axis : ndarray
        Estimated axis of the hole.
    """
    # Estimate axis rotations from segment locations
    axis = get_segmentation_pca(seg_img, seg_vals)
    if np.dot(axis, reference_axis) < 0:  # pragma: no cover
        axis *= -1
    return axis


def estimate_hole_axes_from_segmentation_by_orientation(
    seg_img,
    seg_val_dict,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
):
    """
    Estimate hole axes from segmentation by orientation.

    Parameters
    ----------
    seg_img : SimpleITK.Image
        Segmentation image.
    seg_val_dict : dict
        Nested dictionary of segmentation values for each orientation and AP.
    orient_axes_dict : dict
        Nested dictionary of LPS vectors for each orientation and AP.
    orient_names : list of str
        List of orientation names.
    ap_names : list of str
        List of anterior-posterior names.

    Returns
    -------
    initial_axes : dict
        Dictionary of initial axes as a 3 element vector for each orientation.
    """
    initial_axes = dict()
    for orient in orient_names:
        seg_vals = [
            seg_val_dict[orient][ap]
            for ap in ap_names
            if ap in seg_val_dict[orient]
        ]
        initial_axes[orient] = estimate_hole_axis_from_segmentation(
            seg_img, seg_vals, orient_axes_dict[orient]
        )
    return initial_axes


def calculate_centers_of_mass_for_image_and_segmentation(
    img,
    seg_img,
    initial_axes,
    seg_vals_dict,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    slice_seg_thresh=1,
):
    """
    Calculate centers of mass for image and segmentation.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image used for calculating center of mass.
    seg_img : SimpleITK.Image
        The segmentation image used for selecting elements of `img`.
    initial_axes : dict
        Dictionary of initial axes for each orientation.
    seg_vals_dict : dict
        Nested dictionary of segmentation values for each orientation and
        anterior-posterior names.
    orient_lps_vector_dict : dict, optional
        Dictionary of LPS vectors for each orientation.
    orient_names : list of str, optional
        List of orientation names.
    ap_names : list of str, optional
        List of anterior-posterior names.
    slice_seg_thresh : int, optional
        Minimum number of segmentation elements in a slice required for center
        of mass calculation, by default 1.

    Returns
    -------
    coms : dict
        Nested dictionary of Nx3 NDArray centers of mass for each orientation
        and anterior-posterior name.
    """
    coms = dict()
    for orient in orient_names:
        coms[orient] = dict()
        axis_dim = np.nonzero(orient_axes_dict[orient])[0][0]
        r = rot.rotation_matrix_from_vectors(
            initial_axes[orient], orient_axes_dict[orient]
        )
        s = rot.rotation_matrix_to_sitk(r)
        sinv = s.GetInverse()
        seg_img_rs = sv.resample3D(
            seg_img, sinv, interpolator=sitk.sitkNearestNeighbor
        )
        img_rs = sv.resample3D(
            img, sinv, interpolator=sitk.sitkNearestNeighbor
        )
        for ap in ap_names:
            com = slices_centers_of_mass(
                img_rs,
                seg_img_rs,
                axis_dim,
                seg_vals_dict[orient][ap],
                slice_seg_thresh=slice_seg_thresh,
            )
            coms[orient][ap] = (
                r.T @ com.T
            ).T  # rotate com to original location
    return coms


def estimate_axis_rotations_from_centers_of_mass(
    coms,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
):
    """
    Estimate axis rotations from centers of mass.

    Parameters
    ----------
    coms : dict
        Nested dictionary of Nx3 NDArray centers of mass for each orientation
        and anterior-posterior name.
    orient_axes_dict : dict, optional
        Dictionary of LPS vectors for each orientation.
    orient_names : list of str, optional
        List of orientation names.
    ap_names : list of str, optional
        List of anterior-posterior names.

    Returns
    -------
    orient_rotation_matrices : dict
        Dictionary of 3x3 rotation matrices for each orientation.
    axes : dict
        Dictionary of 3 element vector axes for each orientation.
    """
    orient_rotation_matrices = {orient: dict() for orient in orient_names}
    axes = dict()
    for orient in orient_names:
        # center the coms for each segment separately
        # and find the axis for this orientation
        ccoms = []
        for ap in ap_names:
            com = coms[orient][ap]
            m = np.mean(com, axis=0)
            ccoms.append(com - m[np.newaxis, :])
            joined_ccoms = np.concatenate(ccoms, axis=0)
            tmp_axis = ut.get_first_pca_axis(joined_ccoms)
        if np.dot(tmp_axis, orient_axes_dict[orient]) < 0:
            tmp_axis *= -1

        # remove off-axis mean for each segment separately
        coms_deproj_centered = []
        for ap in ap_names:
            com = coms[orient][ap]
            com_proj = ut.vector_rejection(com, tmp_axis)
            proj_m = np.mean(com_proj, axis=0)
            coms_deproj_centered.append(com - proj_m[np.newaxis, :])
        axis = ut.get_first_pca_axis(joined_ccoms)
        if np.dot(axis, orient_axes_dict[orient]) < 0:
            axis *= -1
        R = rot.rotation_matrix_from_vectors(axis, orient_axes_dict[orient])
        axes[orient] = R.T @ orient_axes_dict[orient]
        orient_rotation_matrices[orient] = R
    return orient_rotation_matrices, axes


def find_rotation_to_match_hole_angles(
    img,
    seg_img,
    initial_orient_rotation_matrices,
    axes,
    seg_vals_dict,
    design_centers=def_design_centers,
    orient_axes_dict=def_orient_axes_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    orient_indices=def_orient_indices,
    hole_order=def_hole_order,
    orient_comparison_axis=def_orient_comparison_axes,
    n_iter=10,
):
    """
    Find rotation matrix to match hole angles.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image.
    seg_img : SimpleITK.Image
        The segmentation image.
    initial_orient_rotation_matrices : dict
        Dictionary of initial orientation rotation matrices indexed by
        orientation.
    axes : dict
        Dictionary of axes for each orientation.
    seg_vals_dict : dict
        Dictionary of segmentation values for each orientation and
        anterior-posterior names.
    design_centers : dict
        Dictionary of design centers for each orientation and
        anterior-posterior names.
    orient_axes_dict : dict
        Dictionary of LPS vectors for each orientation.
    orient_names : list of str
        List of orientation names.
    ap_names : list of str
        List of anterior-posterior names.
    orient_indices : dict
        Dictionary of orientation indices.
    hole_order : dict
        Dictionary of hole order (list of AP names) for each orientation.
    orient_comparison_axis : dict
        Dictionary of comparison axes for each orientation.
    n_iter : int, optional
        Number of iterations, by default 10.

    Returns
    -------
    R : ndarray
        Rotation matrix.
    offsets : ndarray
        Offsets for each hole.
    """
    nhole = np.prod([len(x) for x in (orient_names, ap_names)])
    # Start measuring hole location and orientation using the estimated set of
    # axes
    bases = np.zeros((3, 3))
    bases[:, 1] = axes["horizontal"]  # P
    bases[:, 2] = ut.norm_vec(
        ut.vector_rejection(axes["vertical"], bases[:, 1])
    )  # S
    bases[:, 0] = np.cross(bases[:, 1], bases[:, 2])  # L

    srot = (
        initial_orient_rotation_matrices["horizontal"] @ bases[:, 2]
    )  # rotated S axis
    rad = ut.signed_angle_rh(
        srot,
        lps_axes["dv"],
        lps_axes["ap"],
    )
    Rot_y = Rotation.from_rotvec(rad * lps_axes["ap"])
    R_y = Rot_y.as_matrix()
    R = R_y @ initial_orient_rotation_matrices["horizontal"]

    Sinit = rot.rotation_matrix_to_sitk(R)
    Sinit_inv = Sinit.GetInverse()

    seg_img_current = sv.resample3D(
        seg_img, Sinit_inv, interpolator=sitk.sitkNearestNeighbor
    )
    img_current = sv.resample3D(
        img, Sinit_inv, interpolator=sitk.sitkNearestNeighbor
    )

    found_centers = find_holes_by_orientation(
        img_current,
        seg_img_current,
        seg_vals_dict,
        orient_indices,
        orient_names,
        ap_names,
    )
    found_centers_ang, design_centers_ang = [
        find_hole_angles(
            centers,
            hole_order,
            orient_comparison_axis,
            orient_axes_dict,
            orient_names,
        )
        for centers in (found_centers, design_centers)
    ]
    hole_diffs = np.zeros((nhole, 3))
    for i, (orient, ap) in enumerate(itr.product(orient_names, ap_names)):
        hole_diffs[i, :] = (
            design_centers[orient][ap] - found_centers[orient][ap]
        )
    offsets = np.nanmean(hole_diffs, axis=0)

    iter_angle_err = np.zeros((n_iter + 1, 2))
    iter_hole_diff_err = np.zeros((n_iter + 1, nhole, 3))
    found_centers_curr = found_centers
    found_centers_ang_curr = found_centers_ang

    for iterno in range(n_iter):
        iter_angle_err[iterno, :] = [
            found_centers_ang_curr[orient] - design_centers_ang[orient]
            for orient in orient_names
        ]
        iter_hole_diff_err[iterno, :, :] = offsets[np.newaxis, :] - hole_diffs
        for orient in orient_names:
            ang_err = (
                design_centers_ang[orient] - found_centers_ang_curr[orient]
            )
            Rot_update = Rotation.from_rotvec(
                ang_err * orient_axes_dict[orient]
            )
            R_update = Rot_update.as_matrix()
            R = R_update @ R
            S = rot.rotation_matrix_to_sitk(R)
            S_inv = S.GetInverse()
            seg_img_current = sv.resample3D(
                seg_img, S_inv, interpolator=sitk.sitkNearestNeighbor
            )
            img_current = sv.resample3D(
                img, S_inv, interpolator=sitk.sitkNearestNeighbor
            )
            found_centers_curr = find_holes_by_orientation(
                img_current,
                seg_img_current,
                seg_vals_dict,
                orient_indices,
                orient_names,
                ap_names,
            )
            found_centers_ang_curr = find_hole_angles(
                found_centers_curr,
                hole_order,
                orient_comparison_axis,
                orient_axes_dict,
                orient_names,
            )
            hole_diffs = np.zeros((nhole, 3))
            for i, (orient, ap) in enumerate(
                itr.product(orient_names, ap_names)
            ):
                hole_diffs[i, :] = (
                    design_centers[orient][ap] - found_centers_curr[orient][ap]
                )
            offsets = np.nanmean(hole_diffs, axis=0)
    iter_angle_err[n_iter, :] = [
        found_centers_ang_curr[orient] - design_centers_ang[orient]
        for orient in orient_names
    ]
    iter_hole_diff_err[n_iter, :, :] = offsets[np.newaxis, :] - hole_diffs
    return R, offsets


def estimate_coms_from_image_and_segmentation(
    img,
    seg_img,
    seg_vals_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    orient_axes_dict=def_orient_axes_dict,
):
    """
    Estimate centers of mass (COMs) from image and segmentation.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image.
    seg_img : SimpleITK.Image
        The segmentation image.
    seg_vals_dict : dict
        Nested dictionary of segmentation values.
    orient_names : tuple of str, optional
        Tuple of orientation names, by default ("vertical", "horizontal").
    ap_names : tuple of str, optional
        Tuple of anterior-posterior names, by default
        ("anterior", "posterior").
    lps_axes : dict, optional
        Dictionary of LPS axes, by default dict(ap=np.array([0, 1, 0]),
        dv=np.array([0, 0, 1]), ml=np.array([1, 0, 0])).
    hole_orient_axis : dict, optional
        Dictionary of hole orientation axes, by default dict(horizontal="ap",
        vertical="dv").

    Returns
    -------
    coms : dict
        Dictionary of Nx3 NDArrays of centers of mass for each orientation.
    """
    # Estimate axis rotations from segment locations
    initial_axes = estimate_hole_axes_from_segmentation_by_orientation(
        seg_img, seg_vals_dict, orient_axes_dict, orient_names, ap_names
    )

    # Find centers of mass (COM) for slices perpendicular to intial axes
    coms = calculate_centers_of_mass_for_image_and_segmentation(
        img,
        seg_img,
        initial_axes,
        seg_vals_dict,
        orient_axes_dict,
        orient_names,
        ap_names,
    )
    return coms


def estimate_rotation_and_coms_from_image_and_segmentation(
    img,
    seg_img,
    seg_vals_dict,
    orient_names=def_orient_names,
    ap_names=def_ap_names,
    orient_comparison_axis=def_orient_comparison_axes,
    design_centers=def_design_centers,  # Bregma-relative mms (LPS)
    orient_indices=def_orient_indices,
    hole_order=def_hole_order,
    orient_axes_dict=def_orient_axes_dict,
    n_iter=10,
):
    """
    Estimate rotation and centers of mass (COMs) from image and segmentation.

    Parameters
    ----------
    img : SimpleITK.Image
        The original image.
    seg_img : SimpleITK.Image
        The segmentation image.
    seg_vals_dict : dict
        Dictionary of segmentation values for each orientation and
        anterior-posterior names.
    orient_names : list of str, optional
        List of orientation names, by default `def_orient_names`.
    ap_names : list of str, optional
        List of anterior-posterior names, by default `def_ap_names`.
    orient_comparison_axis : dict, optional
        Dictionary of comparison axes for each orientation, by default
        `def_orient_comparison_axes`.
    design_centers : dict, optional
        Dictionary of design centers for each orientation and
        anterior-posterior names, by default `def_design_centers`.
    orient_indices : dict, optional
        Dictionary of orientation indices, by default `def_orient_indices`.
    hole_order : dict, optional
        Dictionary of hole order for each orientation, by default
        `def_hole_order`.
    orient_axes_dict : dict, optional
        Dictionary of axes for each orientation, by default
        `def_orient_axes_dict`.
    n_iter : int, optional
        Number of iterations, by default 10.

    Returns
    -------
    coms : dict
        Dictionary of centers of mass for each orientation and
        anterior-posterior name.
    R : ndarray
        Rotation matrix.
    offsets : ndarray
        Offsets for each hole.
    """

    coms = estimate_coms_from_image_and_segmentation(
        img,
        seg_img,
        seg_vals_dict,
        orient_names=orient_names,
        ap_names=ap_names,
        orient_axes_dict=orient_axes_dict,
    )

    # Estimate axis rotations from centers of mass
    (
        orient_rotation_matrices,
        axes,
    ) = estimate_axis_rotations_from_centers_of_mass(
        coms,
        orient_axes_dict=orient_axes_dict,
        orient_names=orient_names,
        ap_names=ap_names,
    )

    R, offsets = find_rotation_to_match_hole_angles(
        img,
        seg_img,
        orient_rotation_matrices,
        axes,
        seg_vals_dict,
        design_centers=design_centers,
        orient_axes_dict=orient_axes_dict,
        orient_names=orient_names,
        ap_names=ap_names,
        orient_indices=orient_indices,
        hole_order=hole_order,
        orient_comparison_axis=orient_comparison_axis,
        n_iter=n_iter,
    )

    return coms, R, offsets



import os
import unittest
from pathlib import Path

import numpy as np

from aind_mri_utils.file_io import simpleitk as si
from aind_mri_utils.optimization import create_rigid_transform


class SITKTest(unittest.TestCase):
    """Tests functions in `file_io.simpleitk`."""

    def test_save_sitk_transform(self) -> None:
        """
        Tests that the `save_sitk_transform` function works as intended.
        """
        # Test that a 4x3 transform is created correctly
        trans = create_rigid_transform(0, 0, 0, 0, 0, 0)
        self.assertTrue(np.array_equal(trans[:3, :3], np.eye(3)))
        si.save_sitk_transform(str(Path("testsave.h5")), trans)
        load_trans = si.load_sitk_transform(str(Path("testsave.h5")))
        self.assertTrue(self.close_enough(trans, load_trans))

        # Test that the above can work from 6x1
        si.save_sitk_transform(str(Path("testsave.h5")), [0, 0, 0, 0, 0, 0])
        load_trans = si.load_sitk_transform(str(Path("testsave.h5")))
        self.assertTrue(self.close_enough(trans, load_trans))

        # Test more complicated transform
        trans[0, -1] = 1
        si.save_sitk_transform(str(Path("testsave.h5")), trans)
        load_trans = si.load_sitk_transform(str(Path("testsave.h5")))
        self.assertTrue(self.close_enough(trans, load_trans))

        # Test inversion functionality when saving
        si.save_sitk_transform(
            str(Path("testsave.h5")), trans, transpose_matrix=True
        )
        load_trans = si.load_sitk_transform(str(Path("testsave.h5")))
        load_trans[:3, :3] = load_trans[:3, :3].T

        self.assertTrue(np.array_equal(trans, load_trans))

        # Test inversion functionality when loading
        si.save_sitk_transform(
            str(Path("testsave.h5")), trans, transpose_matrix=True
        )
        load_trans = si.load_sitk_transform(
            str(Path("testsave.h5")), transpose_matrix=True
        )

        self.assertTrue(self.close_enough(trans, load_trans))

        # Test failure modes
        trans = np.eye(4)
        try:
            si.save_sitk_transform(str(Path("testsave.h5")), trans)
        except NotImplementedError:
            self.assertTrue(True)

        try:
            load_trans = si.load_sitk_transform(
                str(Path("testsave.h5")), outshape=(4, 4)
            )
        except NotImplementedError:
            self.assertTrue(True)

        # Kill the file we created- it was just a test
        os.remove("testsave.h5")

    def close_enough(self, a, b):
        """Helper function to check if two arrays are close enough."""
        return np.all(np.abs(a - b) < 0.0001)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Functions for optimizing volume fits.

"""
import numpy as np
from scipy.optimize import fmin

from .measurement import dist_point_to_line, dist_point_to_plane


def append_ones_column(data):
    """
    Add a column of ones to the end of a matrix.

    Parameters
    ----------
    data : np.array
        Data to append a column of ones to.

    Returns
    -------
    np.array
        Data with a column of ones appended to the end.
    """
    return np.hstack([data, np.ones([data.shape[0], 1])])


def create_rigid_transform(rx, ry, rz, cx, cy, cz):
    """
    Create a rigid transform matrix from the given parameters.

    Parameters
    ----------
    rx : float
        x rotation.
    ry : float
        y rotation.
    rz : float
        z rotation.
    cx : float
        x offset.
    cy : float
        y offset.
    cz : float
        z offset.

    Returns
    -------
    R : np.array(4,3)
        Rotation matrix (rigid transform).

    """

    RZ = np.array(
        [
            [np.cos(np.deg2rad(rz)), -np.sin(np.deg2rad(rz)), 0],
            [np.sin(np.deg2rad(rz)), np.cos(np.deg2rad(rz)), 0],
            [0, 0, 1],
        ]
    )
    RX = np.array(
        [
            [1, 0, 0],
            [0, np.cos(np.deg2rad(rx)), -np.sin(np.deg2rad(rx))],
            [0, np.sin(np.deg2rad(rx)), np.cos(np.deg2rad(rx))],
        ]
    )
    RY = np.array(
        [
            [np.cos(np.deg2rad(ry)), 0, np.sin(np.deg2rad(ry))],
            [0, 1, 0],
            [-np.sin(np.deg2rad(ry)), 0, np.cos(np.deg2rad(ry))],
        ]
    )

    R = np.dot(np.dot(RX, RY), RZ)
    R = np.vstack([R, np.array([cx, cy, cz])])
    return R


def cost_function_weighted_labeled_lines(
    T, pts1, pts2, moving, labels, weights
):
    """
    Cost function for optimizing a rigid transform on weighted points.

    Parameters
    ----------
    T : np.array(6,)
        Rigid transform parameters.
    pts1 : np.array(N,3)
        Point on line
    pts2 : np.array(N,3)
        Second point on line
    moving : np.array(M,3)
        Position of points to be transformed.
    labels : np.array(M,dtype=int)
        Labels of points, corresponding to index in
        pts1,pts2,and pts_for_line.
    weights : np.array(M,)
        Weights of points

    Returns
    -------
    Distance
        Sum of distances between points all points and
        their corresponding lines, weighted by weights.

    """
    rx = T[0]
    ry = T[1]
    rz = T[2]
    cx = T[3]
    cy = T[4]
    cz = T[5]
    trans = create_rigid_transform(rx, ry, rz, cx, cy, cz)
    transformed = np.dot(moving, trans)

    D = np.zeros((moving.shape[0], 1))
    for ii in range(pts1.shape[0]):
        lst = np.where(labels == ii)[0]
        for jj in range(len(lst)):
            D[lst[jj]] = (
                dist_point_to_line(
                    pts1[ii, :], pts2[ii, :], transformed[lst[jj], :]
                )
                * weights[lst[jj]]
            )

    return np.sum(D)


def cost_function_weighted_labeled_lines_with_plane(
    T, pts1, pts2, pts_for_line, moving, labels, weights
):
    """
    Cost function for optimizing a rigid transform on weighted points;
    includes labeled lines and labeled planes.

    Parameters
    ----------
    T : np.array(6,)
        Rigid transform parameters.
    pts1 : np.array(N,3)
        Point on line
    pts2 : np.array(N,3)
        Second point on line
    pts_for_line : List[bool] or np.array(N,dtype=bool)
        True if the line should be used for distance calculation,
        False if the plane should be used.
    moving : np.array(M,3)
        position of points to be transformed.
    labels : np.array(M,dtye=int)
        Labels of points, corresponding to index in pts1,pts2,and pts_for_line.
    weights : np.array(M,)

    Returns
    -------
    Distance
        Sum of distances between points all points and
        their corresponding line or plane, weighted by weights.
    """
    rx = T[0]
    ry = T[1]
    rz = T[2]
    cx = T[3]
    cy = T[4]
    cz = T[5]
    trans = create_rigid_transform(rx, ry, rz, cx, cy, cz)
    transformed = np.dot(moving, trans)

    D = np.zeros((moving.shape[0], 1))
    for ii in range(pts1.shape[0]):
        lst = np.where(labels == ii)[0]
        if pts_for_line[ii]:
            for jj in range(len(lst)):
                D[lst[jj]] = (
                    dist_point_to_line(
                        pts1[ii, :], pts2[ii, :], transformed[lst[jj], :]
                    )
                    * weights[lst[jj]]
                )
        else:
            for jj in range(len(lst)):
                D[lst[jj]] = (
                    dist_point_to_plane(
                        pts1[ii, :], pts2[ii, :], transformed[lst[jj], :]
                    )
                    * weights[lst[jj]]
                )

    return np.sum(D)


def _preprocess_weights(weights, positions, normalize, gamma):
    """
    Preprocess weights for use in optimization functions
    """
    if weights is None:
        weights = np.ones((positions.shape[0], 1))
    else:
        # Gamma correct
        # Taken from skimage.exposure.adjust_gamma.
        # Implementing here to avoid importing skimage.
        scale = np.max(weights) - np.min(weights)
        if abs(scale) > 1e-6:
            if gamma is not None:
                if abs(scale) > 1e-6:
                    weights = ((weights / scale) ** gamma) * scale

            if normalize:
                weights = (weights - np.min(weights)) / (scale)
    return weights


def optimize_transform_labeled_lines(
    init,
    pts1,
    pts2,
    positions,
    labels,
    weights=None,
    xtol=1e-12,
    maxfun=10000,
    normalize=False,
    gamma=None,
):
    """
    Function for optimizing a rigid transform on
    weighted points by minimizing distance
    from each point to a specified line.
    Multiple lines can be specified by using labels.

    Parameters
    ----------
    init : np.array(6,1)
        Initial transform, as 6x1 matrix.
    pts1 : np.array(N,3)
        First point on each line
    pts2 : np.array(N,3)
        Second point on each line.
    positions : np.array(M,3)
        Positions of points to optimize on.
    labels : np.array(M,1)
        Labels for each point in positions,
        specifying which line that point too.
    weights : np.array(M,1), optional
        Weights for each point in positions.
        If None is passed, assumes all wieghts are 1.
        Default is None.
    xtol : float, optional
        Stopping tolerence for optimizer. The default is 1e-12.
    maxfun : int, optional
        Max number of function calls for optimizer.
        The default is 10000.
    normalize : bool, optional
        If True, normalize weights to be between 0 and 1.
        The default is False.
    gamma : float, optional
        If value is passed, weight gamma corrected for that value.
        The default is None.

    Returns
    -------
    trans: np.array(4,4)
        Rigid transform matrix that minimizes the cost function.
    T: np.array(6,1)
        Parameters of the rigid transform matrix
        that minimizes the cost function.
    output: tuple
        Fitting data from scipy.optimize.fmin
        (see retol in scipy.optimize.fmin documentation)

    """

    weights = _preprocess_weights(weights, positions, normalize, gamma)

    output = fmin(
        cost_function_weighted_labeled_lines,
        init,
        args=(pts1, pts2, positions, labels, weights),
        xtol=xtol,
        maxfun=maxfun,
    )

    Tframe = output
    print(output)
    trans = create_rigid_transform(
        Tframe[0], Tframe[1], Tframe[2], Tframe[3], Tframe[4], Tframe[5]
    )
    return trans, Tframe


def optimize_transform_labeled_lines_with_plane(
    init,
    pts1,
    pts2,
    pts_for_line,
    positions,
    labels,
    weights=None,
    xtol=1e-12,
    maxfun=10000,
    normalize=False,
    gamma=None,
):
    """
    Function for optimizing a rigid transform on
    weighted points by minimizing distance
    between each point and a specified line or plane.
    Multiple lines/planes can be specified by using labels.

    Parameters
    ----------
    init : np.array(6,1)
        Initial transform, as 6x1 matrix.
    pts1 : np.array(N,3)
        First point on each line OR plane normal
    pts2 : np.array(N,3)
        Second point on each line OR point on plane
    pts_for_line : np.array(N,1)
        Boolean array specifying if each line is a line or a plane.
    positions : np.array(M,3)
        Positions of points to optimize on.
    labels : np.array(M,dtype=np.int))
        Labels for each point in positions,
        specifying which line that point too.
    weights : np.array(M,1), optional
        Weights for each point in positions.
        If None is passed, assumes all wieghts are 1.
        Default is None.
    xtol : float, optional
        Stopping tolerence for optimizer. The default is 1e-12.
    maxfun : int, optional
        Max number of function calls for optimizer.
        The default is 10000.
    normalize : bool, optional
        If True, normalize weights to be between 0 and 1.
        The default is False.
    gamma : float, optional
        If value is passed, weight gamma corrected for that value.
        The default is None.

    Returns
    -------
    trans: np.array(4,4)
        Rigid transform matrix that minimizes the cost function.
    T: np.array(6,1)
        Parameters of the rigid transform matrix
        that minimizes the cost function.
    """
    weights = _preprocess_weights(weights, positions, normalize, gamma)

    output_a = fmin(
        cost_function_weighted_labeled_lines,
        init,
        args=(pts1, pts2, positions, labels, weights),
        xtol=xtol,
        maxfun=maxfun,
    )

    output_b = fmin(
        cost_function_weighted_labeled_lines_with_plane,
        output_a,
        args=(pts1, pts2, pts_for_line, positions, labels, weights),
        xtol=xtol,
        maxfun=maxfun,
    )

    Tframe = output_b
    trans = create_rigid_transform(
        Tframe[0], Tframe[1], Tframe[2], Tframe[3], Tframe[4], Tframe[5]
    )
    return trans, Tframe


def get_headframe_hole_lines(
    version=0.1,
    insert_underscores=False,
    coordinate_system="LPS",
    return_plane=False,
):
    """
    Return the lines for the headframe holes,
    in a format that can be used by the cost function.

    Parameters
    ----------
    version : float, optional
        headframe version to match.
        The default is 0.1.,
        which corresponds to the first-gen zircona hole hemisphere headframe.
    insert_underscores : bool, optional
        If true, insert underscores into the names of the lines.
        The default is False.
    coordinate_system : str, optional
        Coordinate system to return the lines in. The default is 'LPS'.
    return_plane : bool, optional
        Return point for horizontal plane as last point in list
        Default is False

    Returns
    -------
    pts1 : np.array
        One point on each line, in headframe coordinates.
    pts2 : np.array
        Another point on each line.
    names : list
        Name of each line.
    """
    if version == 0.1:
        if coordinate_system == "RAS":
            # Idealized Headframe holes
            ant_vert_hole_pts = np.array(
                [[5.1, -3.2, -1], [5.1, -3.2, 4]]
            )  # Z doesn't matter, just needs two points
            post_vert_hole_pts = np.array(
                [[6.85, -9.9, 0], [6.85, -9.9, 5]]
            )  # Z doesn't matter, just needs two points
            ant_hrz_hole_pts = np.array(
                [[6.34, 0, 2.5], [6.34, -6.5, 2.5]]
            )  # Y doesn't matter, just needs two points
            post_hrz_hole_pts = np.array(
                [[5.04, -6.5, 1], [5.04, -12, 1]]
            )  # Y doesn't matter, just needs two points
        elif coordinate_system == "LPS":
            ant_vert_hole_pts = np.array(
                [[-5.1, 3.2, -1], [-5.1, 3.2, 4]]
            )  # Z doesn't matter, just needs two points
            post_vert_hole_pts = np.array(
                [[-6.85, 9.9, 0], [-6.85, 9.9, 5]]
            )  # Z doesn't matter, just needs two points
            ant_hrz_hole_pts = np.array(
                [[-6.34, 0, 2.5], [-6.34, 6.5, 2.5]]
            )  # Y doesn't matter, just needs two points
            post_hrz_hole_pts = np.array(
                [[-5.04, 6.5, 1], [-5.04, 12, 1]]
            )  # Y doesn't matter, just needs two points

        if insert_underscores:
            names = [
                "anterior_horizontal",
                "anterior_vertical",
                "posterior_horizontal",
                "posterior_vertical",
            ]
        else:
            names = [
                "anterior horizontal",
                "anterior vertical",
                "posterior horizontal",
                "posterior vertical",
            ]

        if return_plane:
            # Last point is point on plane
            pts1 = np.array(
                [
                    ant_hrz_hole_pts[0, :],
                    ant_vert_hole_pts[0, :],
                    post_hrz_hole_pts[0, :],
                    post_vert_hole_pts[0, :],
                    [0, 0, 2],
                ],
            )
            # Last point is plane normal
            pts2 = np.array(
                [
                    ant_hrz_hole_pts[1, :],
                    ant_vert_hole_pts[1, :],
                    post_hrz_hole_pts[1, :],
                    post_vert_hole_pts[1, :],
                    [0, 0, 1],
                ]
            )
            names.append("plane")
        else:
            # Stack in a specific order
            pts1 = np.array(
                [
                    ant_hrz_hole_pts[0, :],
                    ant_vert_hole_pts[0, :],
                    post_hrz_hole_pts[0, :],
                    post_vert_hole_pts[0, :],
                ]
            )
            pts2 = np.array(
                [
                    ant_hrz_hole_pts[1, :],
                    ant_vert_hole_pts[1, :],
                    post_hrz_hole_pts[1, :],
                    post_vert_hole_pts[1, :],
                ]
            )

        return pts1, pts2, names
    else:
        raise ValueError("Version not supported")



"""
Functions for saving and loading transforms using SimpleITK.
"""

import numpy as np
import SimpleITK as sitk

from ..optimization import create_rigid_transform


def save_sitk_transform(filename, T, transpose_matrix=False):
    """
    Save transform to a sitk file (readable by Slicer);

    Parameters
    ----------
    filename : string
        filename to save to.
    T : np.array(4,3), np.array(4x4), or np.array(6,)
        Rigid transform t
        o save.
        if a np.array(6,) is passed, a rigid transform is created using
        aind_mri_utils.optimization.crate_rigid_transform.
    invert = bool, optional
        If true, invert the transform before saving.
        Default is False.


    """
    if len(T) == 6:
        trans = create_rigid_transform(T[0], T[1], T[2], T[3], T[4], T[5])

    elif T.shape == (4, 3):
        trans = T

    else:
        NotImplementedError("T must be a 4x3 or 6x1 array")
        return None

    A = sitk.AffineTransform(3)
    if transpose_matrix:
        A.SetMatrix(trans[:3, :3].T.flatten())
    else:
        A.SetMatrix(trans[:3, :3].flatten())
    A.SetTranslation(trans[3, :])
    sitk.WriteTransform(A, filename)


def load_sitk_transform(filename, transpose_matrix=False, outshape=(4, 3)):
    """
    Convert a sitk transform file to a 4x3 numpy array.

    Parameters
    ----------
    filename : string
        filename to load from.
    invert = bool, optional
        If true, invert the transform before saving.
        Default is False.

    Returns
    -------
    trans: np.array(4,3)
        Affine transform.

    """
    A = sitk.ReadTransform(filename)

    if outshape == (4, 3):
        matrix = np.array(A.GetParameters())[:9].reshape((3, 3))
        if transpose_matrix:
            matrix = matrix.T
        offset = np.array(A.GetParameters()[-3:])
        trans = np.vstack([matrix, offset])
        return trans

    else:
        NotImplementedError("outshape must be (4,3)")
